function check_shape() {
  for (var t = 1; t < arguments.length; t++) {
    var current_tensor = t < 0 || arguments.length <= t ? undefined : arguments[t];
    var previous_tensor = t - 1 < 0 || arguments.length <= t - 1 ? undefined : arguments[t - 1];
    if (current_tensor.shape.length != previous_tensor.shape.length) throw new Error("tensors are in different shape");

    for (var s = 0; s < current_tensor.shape.length; s++) {
      if (current_tensor.shape[s] != previous_tensor.shape[s]) throw new Error("tensors are in different shape");
    }
  }
}
function isTensor(t) {
  return t.data !== undefined;
}
function isVertex(t) {
  return t.grad_ !== undefined;
}

function _typeof(obj) {
  "@babel/helpers - typeof";

  if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") {
    _typeof = function (obj) {
      return typeof obj;
    };
  } else {
    _typeof = function (obj) {
      return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
    };
  }

  return _typeof(obj);
}

function _classCallCheck(instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
}

function _defineProperties(target, props) {
  for (var i = 0; i < props.length; i++) {
    var descriptor = props[i];
    descriptor.enumerable = descriptor.enumerable || false;
    descriptor.configurable = true;
    if ("value" in descriptor) descriptor.writable = true;
    Object.defineProperty(target, descriptor.key, descriptor);
  }
}

function _createClass(Constructor, protoProps, staticProps) {
  if (protoProps) _defineProperties(Constructor.prototype, protoProps);
  if (staticProps) _defineProperties(Constructor, staticProps);
  return Constructor;
}

var Tensor = /*#__PURE__*/function () {
  function Tensor(array, shape) {
    _classCallCheck(this, Tensor);

    this.data = this.extract(array);
    this.shape = shape ? Array.from(shape) : this.findshape(array);

    if (this.data.length != (this.shape.length == 0 ? 1 : this.shape.reduce(function (a, b) {
      return a * b;
    }))) {
      throw new Error("given shape is wrong");
    }
  }

  _createClass(Tensor, [{
    key: "findshape",
    value: function findshape(array) {
      if (Array.isArray(array)) {
        var shape = [];
        shape.push(array.length);
        this.findshape(array[0]).forEach(function (v) {
          shape.push(v);
        });
        return shape;
      } else {
        return [];
      }
    }
  }, {
    key: "extract",
    value: function extract(array) {
      var _this = this;

      if (typeof array != "number") {
        var _ret = function () {
          var res = [];

          for (var a = 0; a < array.length; a++) {
            var el = array[a];

            _this.extract(el).forEach(function (e) {
              res.push(e);
            });
          }

          return {
            v: res
          };
        }();

        if (_typeof(_ret) === "object") return _ret.v;
      } else {
        return [array];
      }
    }
  }, {
    key: "build",
    value: function build() {
      var shape = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : this.shape;
      var data = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : this.data;
      var res = [];

      for (var s = 0; s < shape[0]; s++) {
        res.push(shape.length == 1 ? data[s] : this.build(shape.slice(1), data.slice(s * Math.floor(data.length / shape[0]), s * Math.floor(data.length / shape[0]) + shape.slice(1).reduce(function (a, b) {
          return a * b;
        }))));
      }

      return res;
    }
  }, {
    key: "print",
    value: function print() {
      console.log(this.constructor.name);

      if (this.shape.length == 0) {
        console.log(JSON.stringify(this.data[0]));
        console.log(' ');
        return true;
      }

      var res = JSON.stringify(this.build());
      var newstr = '';
      var count = 0;

      for (var cf = 0; cf < res.length; cf++) {
        if (res[cf] == '[') {
          count++;
        } else {
          break;
        }
      }

      for (var c = 0; c < res.length; c++) {
        if (res[c] == ',' && res[c + 1] == '[') {
          var cn_count = 0;

          for (var cn = c + 1; cn < res.length; cn++) {
            if (res[cn] != '[') break;
            cn_count += 1;
          }

          newstr += '\n'.repeat(cn_count) + ' '.repeat(count - cn_count);
          continue;
        }

        newstr += res[c] == ',' ? ' ' : res[c];
      }

      console.log(newstr);
      console.log(' ');
      return true;
    }
  }]);

  return Tensor;
}();

function add(a, b) {
  check_shape(a, b);
  var res = [];

  for (var el = 0; el < a.data.length; el++) {
    res[el] = a.data[el] + b.data[el];
  }

  return new Tensor(res, a.shape);
}
function sub(a, b) {
  check_shape(a, b);
  var res = [];

  for (var el = 0; el < a.data.length; el++) {
    res[el] = a.data[el] - b.data[el];
  }

  return new Tensor(res, a.shape);
}
function multiply(a, b) {
  check_shape(a, b);
  var res = [];

  for (var el = 0; el < a.data.length; el++) {
    res[el] = a.data[el] * b.data[el];
  }

  return new Tensor(res, a.shape);
}

function _cstep(shape) {
  if (shape.length > 1) {
    var current_shape = [shape.slice(1).reduce(function (a, b) {
      return a * b;
    })];
    return current_shape.concat(_cstep(shape.slice(1)));
  } else {
    return [1];
  }
}

function _cstep_change(step, dimension) {
  if (dimension.length > 0) {
    var currentstate = [step[dimension[0]]];
    return currentstate.concat(_cstep_change(step, dimension.slice(1)));
  } else {
    return [];
  }
}

function _cindex(index, step) {
  if (index.length > 0) {
    var current_index = index[0] * step[0];
    return current_index + _cindex(index.slice(1), step.slice(1));
  } else {
    return 0;
  }
}

function _transpose_main(a, dimension, shape, step) {
  var len = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : shape.length;
  var index = arguments.length > 5 && arguments[5] !== undefined ? arguments[5] : [];
  var arr = [];

  for (var s = 0; s < (shape[0] || 0); s++) {
    var ind = index.concat([s]);

    _transpose_main(a, dimension, shape.slice(1), step, len, ind).forEach(function (el) {
      arr.push(el);
    });
  }

  if (index.length == len) {
    return [a[_cindex(index, step)]];
  }

  return arr;
}

function transpose(a, dimension) {
  var dim = dimension ? dimension.length != 0 ? dimension : a.shape.map(function (v, i) {
    return i;
  }).reverse() : a.shape.map(function (v, i) {
    return i;
  }).reverse();

  var res = _transpose_main(a.data, dim, _cstep_change(a.shape, dim), _cstep_change(_cstep(a.shape), dim));

  var shape = _cstep_change(a.shape, dim);

  return new Tensor(res, shape);
} // Mat mul

function _mat_mul_2d(a, a_shape, b, b_shape) {
  var col_step = a_shape[a_shape.length - 1];
  var dim = b_shape.map(function (a, i) {
    return i;
  }).reverse();

  var b_data = _transpose_main(b, dim, _cstep_change(b_shape, dim), _cstep_change(_cstep(b_shape), dim));

  var bShape = _cstep_change(b_shape, dim);

  var res = [];

  for (var r = 0; r < a.length / col_step; r++) {
    var _loop = function _loop(c) {
      var a_m = a.slice(r * col_step, r * col_step + col_step);
      var b_m = b_data.slice(c * col_step, c * col_step + col_step);
      res.push(a_m.map(function (a, i) {
        return a * b_m[i];
      }).reduce(function (a, b) {
        return a + b;
      }));
    };

    for (var c = 0; c < b_shape[b_shape.length - 1]; c++) {
      _loop(c);
    }
  }

  return res;
}

function _split(a, aShape, b, bShape) {
  var a_shape = aShape.slice(aShape.length - 2);
  var b_shape = bShape.slice(bShape.length - 2);
  var tot_el_a = a_shape[0] * a_shape[1];
  var tot_el_b = b_shape[0] * b_shape[1];
  var res = [];

  for (var i = 0; i < a.length / tot_el_a; i++) {
    _mat_mul_2d(a.slice(i * tot_el_a, i * tot_el_a + tot_el_a), a_shape, b.slice(i * tot_el_b, i * tot_el_b + tot_el_b), b_shape).forEach(function (v) {
      return res.push(v);
    });
  }

  return res;
}

function matmul(a, b) {
  if (a.shape.length > 1 && b.shape.length > 1) {
    if (a.shape[a.shape.length - 1] != b.shape[a.shape.length - 2]) throw new Error("number of columns in first tensor should equal to number of rows in second tensor");

    if (a.shape.length > 2) {
      for (var s = 0; s < a.shape.length - 2; s++) {
        if (a.shape[s] != b.shape[s]) throw new Error("two tensor are not equal");
      }
    }
  } else {
    return multiply(a, b);
  }

  var res = _split(a.data, a.shape, b.data, b.shape);

  var shape = Array.from(a.shape);
  shape[shape.length - 1] = b.shape[b.shape.length - 1];
  return new Tensor(res, shape);
}

/**
 * Creates a Tensor with random elements (it uses Math.random() to generate random values).
 * @param shape the output tensor shape.
 * 
 * @returns Tensor.
 */

function rand(shape) {
  var size = shape.reduce(function (a, b) {
    return a * b;
  });
  var res = [];

  for (var s = 0; s < size; s++) {
    res.push(Math.random());
  }

  return new Tensor(res, shape);
}
/**
 * Creates a Tensor with an element zero (0).
 * @param shape the output tensor shape.
 * 
 * @returns Tensor.
 */

function zeros(shape) {
  var size = shape.reduce(function (a, b) {
    return a * b;
  });
  var res = [];

  for (var s = 0; s < size; s++) {
    res.push(0);
  }

  return new Tensor(res, shape);
}
/**
 * Creates a Tensor with an element One (1).
 * @param shape the output tensor shape.
 * 
 * @returns Tensor.
 */

function ones(shape) {
  var size = shape.reduce(function (a, b) {
    return a * b;
  });
  var res = [];

  for (var s = 0; s < size; s++) {
    res.push(1);
  }

  return new Tensor(res, shape);
}
/**
 * Creates a Tensor filled with a value (number).
 * @param shape the output tensor shape.
 * @param value The number to fill the tensor with.
 * 
 * @returns Tensor.
 */

function fill(shape, value) {
  var size = shape.reduce(function (a, b) {
    return a * b;
  });
  var res = [];

  for (var s = 0; s < size; s++) {
    res.push(value);
  }

  return new Tensor(res, shape);
}
/**
 * Creates a Tensor filled with a return value of the method (fn).
 * @param shape The output tensor shape.
 * @param fn The Tensor is filled with a return value of this method.
 * 
 * @returns Tensor.
 */

function fillfn(shape) {
  var fn = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : function () {
    return 1;
  };
  var size = shape.reduce(function (a, b) {
    return a * b;
  });
  var res = [];

  for (var s = 0; s < size; s++) {
    res.push(fn());
  }

  return new Tensor(res, shape);
}

var Tensor$1 = /*#__PURE__*/function () {
  function Tensor(array, shape) {
    _classCallCheck(this, Tensor);

    this.data = this.extract(array);
    this.shape = shape ? Array.from(shape) : this.findshape(array);

    if (this.data.length != (this.shape.length == 0 ? 1 : this.shape.reduce(function (a, b) {
      return a * b;
    }))) {
      throw new Error("given shape is wrong");
    }
  }

  _createClass(Tensor, [{
    key: "findshape",
    value: function findshape(array) {
      if (Array.isArray(array)) {
        var shape = [];
        shape.push(array.length);
        this.findshape(array[0]).forEach(function (v) {
          shape.push(v);
        });
        return shape;
      } else {
        return [];
      }
    }
  }, {
    key: "extract",
    value: function extract(array) {
      var _this = this;

      if (typeof array != "number") {
        var _ret = function () {
          var res = [];

          for (var a = 0; a < array.length; a++) {
            var el = array[a];

            _this.extract(el).forEach(function (e) {
              res.push(e);
            });
          }

          return {
            v: res
          };
        }();

        if (_typeof(_ret) === "object") return _ret.v;
      } else {
        return [array];
      }
    }
  }, {
    key: "build",
    value: function build() {
      var shape = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : this.shape;
      var data = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : this.data;
      var res = [];

      for (var s = 0; s < shape[0]; s++) {
        res.push(shape.length == 1 ? data[s] : this.build(shape.slice(1), data.slice(s * Math.floor(data.length / shape[0]), s * Math.floor(data.length / shape[0]) + shape.slice(1).reduce(function (a, b) {
          return a * b;
        }))));
      }

      return res;
    }
  }, {
    key: "print",
    value: function print() {
      console.log(this.constructor.name);

      if (this.shape.length == 0) {
        console.log(JSON.stringify(this.data[0]));
        console.log(' ');
        return true;
      }

      var res = JSON.stringify(this.build());
      var newstr = '';
      var count = 0;

      for (var cf = 0; cf < res.length; cf++) {
        if (res[cf] == '[') {
          count++;
        } else {
          break;
        }
      }

      for (var c = 0; c < res.length; c++) {
        if (res[c] == ',' && res[c + 1] == '[') {
          var cn_count = 0;

          for (var cn = c + 1; cn < res.length; cn++) {
            if (res[cn] != '[') break;
            cn_count += 1;
          }

          newstr += '\n'.repeat(cn_count) + ' '.repeat(count - cn_count);
          continue;
        }

        newstr += res[c] == ',' ? ' ' : res[c];
      }

      console.log(newstr);
      console.log(' ');
      return true;
    }
  }]);

  return Tensor;
}();
/**
 * Creates a Tensor with the provided values and shape.
 * @param values The values of the tensor. it can be n-dimensional array of numbers, or a TypedArray. 
 * @param shape The shape of the tensor. if it is not defined, it will be automatically found from values.
 */

function tensor(values, shape) {
  return new Tensor$1(values, shape);
}

function apply_activation(a, act) {
  var res = a.data.map(function (v) {
    return act(v);
  });
  var shape = Array.from(a.shape);
  return new Tensor$1(res, shape);
}

var Vertex = /*#__PURE__*/function () {
  function Vertex(tensor_, parents_) {
    var grad_ = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : new Tensor(tensor_.data.map(function (v) {
      return 0;
    }), tensor_.shape);
    var name = arguments.length > 3 ? arguments[3] : undefined;

    _classCallCheck(this, Vertex);

    this.tensor_ = tensor_;
    this.parents_ = parents_;
    this.grad_ = grad_;
    this.name = name;
  }

  _createClass(Vertex, [{
    key: "back",
    value: function back() {}
  }, {
    key: "print",
    value: function print() {
      if (this.name) console.log(this.name);
      this.tensor_.print();
    }
  }, {
    key: "set_grad",
    value: function set_grad(grad) {
      this.grad_ = grad;
    }
  }]);

  return Vertex;
}();

/**
 * Adds two Tensors or Vertex.
 * @param a The first Tensor to add.
 * @param b The Second Tensor to add.
 * 
 * @returns Tensor if inputs are Tensor, Vertex if inputs are Vertex 
 */

function add$1(a, b) {
  if (isVertex(a) && isVertex(b)) {
    var res_tensor = add(a.tensor_, b.tensor_);
    var res = new Vertex(res_tensor, [a, b]);

    res.back = function () {
      a.grad_ = add(a.grad_, res.grad_);
      b.grad_ = add(b.grad_, res.grad_);
    };

    return res;
  } else if (isTensor(a) && isTensor(b)) {
    return add(a, b);
  } else {
    throw new Error("inputs should be same type");
  }
}
/**
 * Subtracts two Tensors or Vertex.
 * @param a The first Tensor to Subtract.
 * @param b The Second Tensor to Subtract.
 * 
 * @returns Tensor if inputs are Tensor, Vertex if inputs are Vertex 
 */

function sub$1(a, b) {
  if (isVertex(a) && isVertex(b)) {
    var res_tensor = sub(a.tensor_, b.tensor_);
    var res = new Vertex(res_tensor, [a, b]);

    res.back = function () {
      a.grad_ = add(a.grad_, res.grad_);
      var res_grad_new = tensor(res.grad_.data.map(function (v) {
        return v * -1;
      }), res.grad_.shape);
      b.grad_ = add(b.grad_, res_grad_new);
    };

    return res;
  } else if (isTensor(a) && isTensor(b)) {
    return sub(a, b);
  } else {
    throw new Error("inputs should be same type");
  }
}
/**
 * Multiplies two Tensors or Vertex.
 * @param a The first Tensor to Multiply.
 * @param b The Second Tensor to Multiply.
 * 
 * @returns Tensor if inputs are Tensor, Vertex if inputs are Vertex 
 */

function multiply$1(a, b) {
  if (isVertex(a) && isVertex(b)) {
    var res_tensor = multiply(a.tensor_, b.tensor_);
    var res = new Vertex(res_tensor, [a, b]);

    res.back = function () {
      a.grad_ = add(a.grad_, multiply(res.grad_, b.tensor_));
      b.grad_ = add(b.grad_, multiply(res.grad_, a.tensor_));
    };

    return res;
  } else if (isTensor(a) && isTensor(b)) {
    return multiply(a, b);
  } else {
    throw new Error("inputs should be same type");
  }
}

/**
 * Transposes the Tensor according to the dimension (dim).
 * 
 * this operation is a regular matrix transpose on Tensor.
 * The input Tensor axis will be changed according to the input dimension (dim). 
 *  
 * @param a The input tensor
 * @param dim the input dimension
 * 
 * @returns Tensor if inputs are Tensor, Vertex if inputs are Vertex 
 * 
 */

function transpose$1(a, dim) {
  if (isVertex(a)) {
    var res_tensor = transpose(a.tensor_, dim);
    var res = new Vertex(res_tensor, [a]);

    res.back = function () {
      a.grad_ = add(a.grad_, transpose(res.grad_, dim));
    };

    return res;
  } else if (isTensor(a)) {
    return transpose(a, dim);
  } else {
    throw new Error("inputs should be same type");
  }
}
/**
 * Computes the matrix multiplication of two Tensors
 * @param a The first Tensor to Matmul.
 * @param b The Second Tensor to Matmul.
 * 
 * @returns Tensor if inputs are Tensor, Vertex if inputs are Vertex 
 * 
 */

function matmul$1(a, b) {
  if (isVertex(a) && isVertex(b)) {
    var res_tensor = matmul(a.tensor_, b.tensor_);
    var res = new Vertex(res_tensor, [a, b]);

    res.back = function () {
      a.grad_ = add(a.grad_, matmul(res.grad_, transpose(b.tensor_)));
      b.grad_ = add(b.grad_, matmul(transpose(a.tensor_), res.grad_));
    };

    return res;
  } else if (isTensor(a) && isTensor(b)) {
    return matmul(a, b);
  } else {
    throw new Error("inputs should be same type");
  }
}

/**
 * This will Compute the gradient (derivatives) of the current vertex's tensor (tensor_) and 
 * adds the results with grad_ (grad_ is initialized with value (0)).
 * 
 * grad_ must be zero before calling it.
 * 
 * The graph which is constructed while the forword operation is differentiated using chain rule. 
 *
 * @param res Resultant vertex or Starting vertex.
 * @param initial_grad Initial grad or derivative to start with.
 */
function backpass(res, initial_grad) {
  if (!isVertex(res)) throw new Error("backpass will only works with vertex");
  if (initial_grad) res.grad_ = initial_grad;
  res.back();

  if (res.parents_) {
    res.parents_.forEach(function (p) {
      backpass(p);
    });
  }
}
/**
 * This will update the tensor with new values (grad_).
 * this should be called after backpass.
 * 
 * @param res Resultant vertex or Starting vertex.
 * @param rate Learning Rate
 */

function update_loss(res) {
  var rate = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 0.04;
  if (!isVertex(res)) throw new Error("update_loss will only works with vertex");

  for (var i = 0; i < res.tensor_.data.length; i++) {
    var val = res.grad_.data[i] * rate;
    res.tensor_.data[i] -= val ? val : 0;
  }

  if (res.parents_) {
    res.parents_.forEach(function (p) {
      update_loss(p, rate);
    });
  }
}
/**
 * This will Reset the grad (grad_).
 * this should be called after update_loss.
 * 
 * @param res Resultant vertex or Starting vertex.
 */

function grad_zero(res) {
  if (!isVertex(res)) throw new Error("grad_zero will only works with vertex");
  res.grad_.data = res.grad_.data.map(function (v) {
    return 0;
  });

  if (res.parents_) {
    res.parents_.forEach(function (p) {
      grad_zero(p);
    });
  }
}
/**
 * This will detach the graph.
 * this should be called at end.
 * 
 * @param res Resultant vertex or Starting vertex.
 */

function detach(res) {
  if (!isVertex(res)) throw new Error("detach will only works with vertex");

  if (res.parents_) {
    res.parents_.forEach(function (p) {
      detach(p);
    });
  }

  if (res.parents_) {
    res.parents_.forEach(function (v, i) {
      if (res.parents_) delete res.parents_[i];
    });
  }
}
/**
 * This method will visits each vertex of the graph and prints the tensor_ or grad_
 * @param res Resultant vertex or Starting vertex.
 * @param key key to print, "tensor_" | "grad_"
 */

function traversal(res) {
  var key = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "tensor_";
  if (!isVertex(res)) throw new Error("traversal will only works with vertex");
  if (!res.parents_) if (res.name) console.log(res.name);
  res[key].print();

  if (res.parents_) {
    res.parents_.forEach(function (p) {
      traversal(p, key);
    });
  }
}

var Vertex$1 = /*#__PURE__*/function () {
  function Vertex(tensor_, parents_) {
    var grad_ = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : new Tensor(tensor_.data.map(function (v) {
      return 0;
    }), tensor_.shape);
    var name = arguments.length > 3 ? arguments[3] : undefined;

    _classCallCheck(this, Vertex);

    this.tensor_ = tensor_;
    this.parents_ = parents_;
    this.grad_ = grad_;
    this.name = name;
  }

  _createClass(Vertex, [{
    key: "back",
    value: function back() {}
  }, {
    key: "print",
    value: function print() {
      if (this.name) console.log(this.name);
      this.tensor_.print();
    }
  }, {
    key: "set_grad",
    value: function set_grad(grad) {
      this.grad_ = grad;
    }
  }]);

  return Vertex;
}();
/**
 * Creates a new vertex with the provided initial tensor.
 * @param tensor Initial value for the tensor.
 * @param parents Parents of the resultant vertex, it will be filled when use the tensor operations.
 * @param grad Derivatives of this resultant tensor, it will be filled while backpass
 * @param name Name of the vertex.
 */

function vertex(tensor, parents, grad, name) {
  return new Vertex$1(tensor, parents, grad, name);
}

function apply_activation$1(a, act, actPrime) {
  if (isVertex(a)) {
    var ten = apply_activation(a.tensor_, act);
    var res = new Vertex(ten, [a]);

    res.back = function () {
      a.grad_ = add(a.grad_, multiply(res.grad_, apply_activation(a.tensor_, actPrime)));
    };

    return res;
  } else if (isTensor(a)) {
    return apply_activation(a, act);
  } else {
    throw new Error("inputs should be same type");
  }
}

module.exports = {
  tensor: tensor,
  vertex: vertex,
  rand: rand,
  ones: ones,
  zeros: zeros,
  fill: fill,
  fillfn: fillfn,
  applyfn: apply_activation$1,
  add: add$1,
  sub: sub$1,
  multiply: multiply$1,
  matmul: matmul$1,
  transpose: transpose$1,
  backpass: backpass,
  update_loss: update_loss,
  grad_zero: grad_zero,
  traversal: traversal,
  detach: detach
};
